package org.nutz.dao.sql;

import java.util.Set;

import org.nutz.lang.Lang;
import org.nutz.lang.Strings;
import org.nutz.lang.util.LinkedIntArray;

/**
 * @author zozoh
 * @author wendal(wendal1985@gmail.com)
 */
public class SqlLiteral implements Cloneable {

	private void reset() {
		params = new SimpleVarSet();
		vars = new SimpleVarSet();
		stack = new WorkingStack();
		varIndexes = new VarIndexImpl();
		paramIndexes = new VarIndexImpl();
		statementIndexes = new VarIndexImpl();
	}

	private VarSet params;
	private VarSet vars;
	private WorkingStack stack;
	private VarIndexImpl varIndexes;
	private VarIndexImpl paramIndexes;
	private VarIndexImpl statementIndexes;
	private String source;

	public VarSet getParams() {
		return params;
	}

	public void setParams(VarSet holders) {
		this.params = holders;
	}

	public VarSet getVars() {
		return vars;
	}

	public void setVars(VarSet vars) {
		this.vars = vars;
	}

	VarIndex getVarIndexes() {
		return varIndexes;
	}

	VarIndex getParamIndexes() {
		return paramIndexes;
	}

	/**
	 * [@|$][a-zA-Z0-9_-.]+
	 * 
	 * <pre>
	 * 48-57	0-9
	 * 65-90	A-Z
	 * 97-122	a-z
	 * 95		_
	 * 45		-
	 * 46		.
	 * </pre>
	 * 
	 */
	public SqlLiteral valueOf(String str) {
		reset();
		int statementIndex = 1;
		source = str;
		if (null == source)
			return this;
		char[] cs = Strings.trim(source).toCharArray();
		StringBuilder sb;
		for (int i = 0; i < cs.length; i++) {
			char c = cs[i];
			switch (c) {
			case '@':
				if (cs[i + 1] == '@') {
					stack.push(c);
					i++;
					break;
				}
				sb = new StringBuilder();
				i = readTokenName(cs, i, sb);
				// Fail to read token name
				if (sb.length() == 0) {
					stack.push(c);
				} else {
					String name = sb.toString();
					paramIndexes.add(name, stack.markToken());
					statementIndexes.add(name, statementIndex++);
				}
				break;
			case '$':
				if (cs[i + 1] == '$') {
					stack.push(c);
					i++;
					break;
				}
				sb = new StringBuilder();
				i = readTokenName(cs, i, sb);
				// Fail to read token name
				if (sb.length() == 0) {
					stack.push(c);
				} else
					varIndexes.add(sb.toString(), stack.markToken());
				break;
			default:
				stack.push(c);
			}
		}
		stack.finish();
		return this;
	}

	private int readTokenName(char[] cs, int i, StringBuilder sb) {
		for (++i; i < cs.length; i++) {
			int b = (int) cs[i];
			// Special case for underline ('_')
			if (b == 95) {
				sb.append((char) b);
			} else if ((b >= 0 && b <= 47)
						|| (b >= 58 && b <= 64)
						|| (b >= 91 && b <= 96)
						|| (b >= 123 && b <= 160)) {
				break;
			} else {
				sb.append((char) b);
			}
		}
		return i - 1;
	}

	public String toPreparedStatementString() {
		String[] ss = stack.cloneChain();
		autoFill(vars, varIndexes, ss);
		fillParams(ss, "?");
		return Lang.concat(ss).toString();
	}

	private void autoFill(VarSet set, VarIndexImpl indexes, String[] ss) {
		for (String name : set.keys()) {
			Object value = set.get(name);
			String vs = null == value ? "" : value.toString();
			int[] is = indexes.get(name);
			if (null != is)
				for (int i : is)
					ss[stack.getIndex(i)] = vs;
		}
	}

	private void fillParams(String[] ss, String v) {
		for (LinkedIntArray lia : paramIndexes.values()) {
			for (int i : lia.toArray())
				ss[stack.getIndex(i)] = v;
		}
	}

	public int[] getParamIndexes(String name) {
		return statementIndexes.get(name);
	}

	@Override
	public SqlLiteral clone() {
		return new SqlLiteral().valueOf(source);
	}

	public String getSource() {
		return source;
	}

	public String toString() {
		String[] ss = stack.cloneChain();
		autoFill(vars, varIndexes, ss);
		// Fill params, considering the type
		autoFill(params, paramIndexes, ss);
		for (String name : params.keys()) {
			Object value = params.get(name);
			String vs;
			if (null == value) {
				vs = "";
			} else if (value instanceof String) {
				vs = "'" + value + "'";
			} else {
				vs = value.toString();
			}
			int[] is = paramIndexes.get(name);
			if (null != is)
				for (int i : is)
					ss[stack.getIndex(i)] = vs;
		}
		// Done
		return Lang.concat(ss).toString();
	}

	public boolean isSELECT() {
		return stack.firstEquals("SELECT") || stack.firstEquals("WITH");
	}

	public boolean isUPDATE() {
		return stack.firstEquals("UPDATE");
	}

	public boolean isINSERT() {
		return stack.firstEquals("INSERT");
	}

	public boolean isDELETE() {
		return stack.firstEquals("DELETE");
	}

	public boolean isCREATE() {
		return stack.firstEquals("CREATE");
	}

	public boolean isDROP() {
		return stack.firstEquals("DROP");
	}

	public boolean isTRUNCATE() {
		return stack.firstEquals("TRUNCATE");
	}
	
	/**
	 * <code>
	 * UPDATE dao_platoon SET name=@name1,base=@baseName2,leader=@leaderName3 WHERE id=@id4
	 * 取得"leaderName3", "id4", "baseName2", "name1" 
	 * </code>
	 * @return 参数列表
	 */
	public String[] getParamNames() {
		VarIndex paramIndexes2 = getParamIndexes();

		return findNames(paramIndexes2);
	}

	/**<code>
	 * InSeRT INTO $T ($id,$name) VALUES(@id1,@name2)
	 * 取得"T", "name", "id"
	 * </code>
	 * @return 表名称等变量列表
	 */
	public String[] getVarNames() {
		return findNames(getVarIndexes());
	}

	/**
	 * @param var
	 * @return
	 */
	private String[] findNames(VarIndex var) {
		if (var == null)
			return null;
		Set<String> names = var.names();
		String[] param = new String[names.size()];
		names.toArray(param);

		return param;
	}

}
